from pyexlatex.texgen import _usepackage_str


class Package:
    r"""
    Represents LaTeX \usepackage{}, pass to Document if any custom LaTeX packages are needed.
    """

    def __init__(self, name, modifier_str=None):
        self.name = name
        self.modifier_str = modifier_str

    def __repr__(self):
        return f'<Package(name={self.name})>'

    def __str__(self):
        return _usepackage_str(self.name, self.modifier_str)

    def __eq__(self, other):
        return (self.name == other.name) and (self.modifier_str == other.modifier_str)

    def __hash__(self):
        return hash((self.name, self.modifier_str))

    def matches_name(self, other):
        return self.name == other




from typing import Sequence, Optional
from pyexlatex.models.item import ItemBase
from pyexlatex.models.section.base import TextAreaMixin
from pyexlatex.presentation.beamer.overlay.overlay import Overlay
from pyexlatex.presentation.beamer.overlay.commands.uncover import Uncover


class TikZItem(TextAreaMixin, ItemBase):

    def __init__(self, name, contents, options: Optional[Sequence[str]] = None,
                 overlay: Optional['Overlay'] = None):
        self.contents = contents
        self.options = options
        self.overlay = overlay
        self.name = name
        super().__init__(name, contents)

    def __str__(self):
        item_str = fr'\{self.name} {self.options_str} {self.contents};'
        if self.overlay:
            item_str = Uncover(item_str, overlay=self.overlay)
        return item_str

    @property
    def options_str(self) -> str:
        if self.options is None:
            return ''
        return self._wrap_with_bracket(', '.join(self.options))






from typing import Optional, TYPE_CHECKING
if TYPE_CHECKING:
    from pyexlatex.graphics.tikz import Node
from pyexlatex.models.item import ItemBase
from pyexlatex.models.package import Package
from pyexlatex.graphics.tikz.library import TikZLibrary


class DirectionBase(ItemBase):
    """
    Base class for relative positioning directions in TikZ, e.g. above, above right, etc.
    """

    name = '<invalid, do not use DirectionBase directly>'
    is_vertical_direction = None

    def __init__(self, direction: Optional['DirectionBase'] = None, of: Optional['Node'] = None,
                 by: Optional[int] = None):
        self.child_direction = direction
        self.of = of
        self.by = by
        self._validate()

        self.init_data()
        self.data.packages.extend([
            Package('tikz'),
            TikZLibrary('positioning')
        ])

    def _validate(self):
        if self.child_direction is not None and self.of is not None:
            raise ValueError('must only pass of when there are no other child directions')
        if self.is_vertical_direction is None:
            raise ValueError('must set is_vertical_direction to True or False when subclassing DirectionBase')
        if self.child_direction is not None:
            if self.is_vertical_direction and self.child_direction.is_vertical_direction:
                raise ValueError('cannot combine vertical directions, must be a vertical and horizontal direction')
            if not self.is_vertical_direction and not self.child_direction.is_vertical_direction:
                raise ValueError('cannot combine horizontal directions, must be a vertical and horizontal direction')
            if self.of is not None and self.child_direction.of is not None:
                raise ValueError('cannot pass two of in nested directions')

    @property
    def full_direction(self) -> str:
        if self.child_direction is None:
            return f'{self.name}{self.of_str}'

        if self.is_vertical_direction:
            vertical_name = self.name
            horizontal_name = self.child_direction.name
        else:
            vertical_name = self.child_direction.name
            horizontal_name = self.name
        return f'{vertical_name} {horizontal_name}{self.of_str}'

    @property
    def of_str(self) -> str:
        if self.of is None and self.child_direction.of is None:
            return ''

        label = None
        if self.of is not None:
            label = self.of.label
        else:
            label = self.child_direction.of.label

        return f'={self.by_str}of {label}'

    @property
    def by_str(self) -> str:
        if self.by is None and (self.child_direction is None or self.child_direction.by is None):
            return ''

        vertical_by, horizontal_by = None, None
        if _is_vertical_direction_and_has_by(self):
            vertical_by = self.by
        elif self.child_direction is not None and _is_vertical_direction_and_has_by(self.child_direction):
            vertical_by = self.child_direction.by

        if _is_horizontal_direction_and_has_by(self):
            horizontal_by = self.by
        elif self.child_direction is not None and _is_horizontal_direction_and_has_by(self.child_direction):
            horizontal_by = self.child_direction.by

        out_str = ''
        if vertical_by is not None:
            out_str += f'{vertical_by} '
        if horizontal_by is not None:
            out_str += f'and {horizontal_by} '

        return out_str

    def __str__(self) -> str:
        return self.full_direction


class VerticalDirection(DirectionBase):
    is_vertical_direction = True


class HorizontalDirection(DirectionBase):
    is_vertical_direction = False


class Above(VerticalDirection):
    """
    Used for placing nodes: construct with one node, pass as placement to another node, that second node will
    be placed above the first node.
    """
    name = 'above'


class Right(HorizontalDirection):
    """
    Used for placing nodes: construct with one node, pass as placement to another node, that second node will
    be placed to the right of the first node.
    """
    name = 'right'


class Below(VerticalDirection):
    """
    Used for placing nodes: construct with one node, pass as placement to another node, that second node will
    be placed below the first node.
    """
    name = 'below'


class Left(HorizontalDirection):
    """
    Used for placing nodes: construct with one node, pass as placement to another node, that second node will
    be placed to the left of the first node.
    """
    name = 'left'


def _is_vertical_direction_and_has_by(direction: 'DirectionBase') -> bool:
    return direction.is_vertical_direction and direction.by is not None


def _is_horizontal_direction_and_has_by(direction: 'DirectionBase') -> bool:
    return not direction.is_vertical_direction and direction.by is not None


from pyexlatex.models.item import SimpleItem
from pyexlatex.logic.builder import _build

class FancyHeader(SimpleItem):
    name = 'fancyhead'

    def __init__(self, contents):
        super().__init__(self.name, contents)

remove_header_line = r'\renewcommand{\headrulewidth}{0pt}'
remove_header = _build([
    remove_header_line,
    FancyHeader('')  # remove header default contents
])


from pyexlatex.models.document import Document
from pyexlatex.models.equation import Equation
from pyexlatex.models.section.paragraphs import Paragraph, SubParagraph
from pyexlatex.models.section.sections import Section, SubSection, SubSubSection
from pyexlatex.models.section.appendix import Appendix
from pyexlatex.models.lists.unordered import UnorderedList
from pyexlatex.models.lists.ordered import OrderedList
from pyexlatex.models.format.raw import Raw
from pyexlatex.models.format.text.text import Text
from pyexlatex.models.ref import Ref
from pyexlatex.models.footnote import Footnote
from pyexlatex.models.references.bibtex.article import BibTexArticle
from pyexlatex.models.references.bibtex.misc import BibTexMisc
from pyexlatex.models.references.bibtex.manual import BibTexManual
from pyexlatex.models.references.bibliography import Bibliography
from pyexlatex.models.references.citations import Cite, CiteP, CiteT
from pyexlatex.models.format.text.color import TextColor
from pyexlatex.models.format.text.monospace import Monospace
from pyexlatex.models.format.text.underline import Underline
from pyexlatex.models.format.text.bold import Bold
from pyexlatex.models.format.text.italics import Italics
from pyexlatex.models.format.code.python import Python
from pyexlatex.figure.models.graphic import Graphic
from pyexlatex.models.format.vfill import VFill
from pyexlatex.models.hyperlinks import Hyperlink
from pyexlatex.models.package import Package



from pyexlatex.texgen import _multi_option_item_str
from pyexlatex.models.item import MultiOptionSimpleItem
from pyexlatex.models.package import Package
from pyexlatex.models.section.base import TextAreaMixin

class TextColor(TextAreaMixin, MultiOptionSimpleItem):
    """
    Changes the color of text.
    """
    name = 'textcolor'

    def __init__(self, content, color: str, opacity: float = 1.0, **kwargs):
        self.color = color
        self.opacity = opacity
        self.init_data()
        self.data.packages.append(Package('xcolor'))
        self.content = content
        contents = self.format_contents(content)
        MultiOptionSimpleItem.__init__(self, self.name, self.color_str, contents, **kwargs)

    @property
    def opacity_str(self) -> str:
        if self.opacity != 1:
            return f'!{self.opacity * 100:.0f}'

        return ''

    @property
    def color_str(self) -> str:
        return f'{self.color}{self.opacity_str}'

    def __str__(self):
        return _multi_option_item_str(self.name, self.color_str, self.content, overlay=self.overlay)



from typing import Sequence, Tuple, Optional, TYPE_CHECKING
if TYPE_CHECKING:
    from pyexlatex.presentation.beamer.overlay import Overlay
from pyexlatex.graphics.tikz.path import SpecificPath


class Rectangle(SpecificPath):
    """
    Draws a rectangle.
    """
    draw_type = 'rectangle'
    path_type = 'draw'

    def __init__(self, width: int, height: int, offset: Tuple[int, int] = (0, 0),
                 options: Optional[Sequence[str]] = None,
                 overlay: Optional['Overlay'] = None):
        self.size = (width, height)
        self.offset = offset
        super().__init__(self.get_points(), options=options, overlay=overlay)

    def get_points(self) -> Tuple[Tuple[int, int], Tuple[int, int]]:
        return (
            self.offset,
            (self.size[0] + self.offset[0], self.size[1] + self.offset[1])
        )


from typing import List, Optional, Sequence
from copy import deepcopy
from pyexlatex.models.document import DocumentBase
from pyexlatex.typing import ItemOrListOfItems
from pyexlatex.models.package import Package
from pyexlatex.models.control.documentclass import DocumentClass
from pyexlatex.models.control.mode import Mode
from pyexlatex.presentation.beamer.theme.usetheme import UseTheme
from pyexlatex.models.title.frame import TitleFrame, should_create_title_frame
from pyexlatex.models.item import ItemBase
from pyexlatex.presentation.beamer.templates.lists.dim_reveal_items import eliminate_dim_reveal


class Presentation(DocumentBase):
    """
    The main high-level class for creating presentations, can render presentations as PDF in either a presenter
    style or a handouts style.
    """
    name = 'document'

    def __init__(self, content: ItemOrListOfItems, packages: List[Package]=None,
                 pre_env_contents: Optional[ItemOrListOfItems] = None,
                 title: Optional[str] = None, author: Optional[str] = None, date: Optional[str] = None,
                 short_title: Optional[str] = None, subtitle: Optional[str] = None, short_author: Optional[str] = None,
                 institutions: Optional[Sequence[Sequence[str]]] = None, short_institution: Optional[str] = None,
                 font_size: Optional[float] = 11, theme: str = 'Madrid', backend: str = 'beamer',
                 handouts: bool = False):

        self.init_data()
        self.title_frame = None

        if isinstance(content, (ItemBase, str)):
            content = [content]
        else:
            content = deepcopy(content)  # don't overwrite existing content

        if backend != 'beamer':
            raise NotImplementedError('only beamer backend is currently supported for Presentation')

        if handouts:
            doc_class_options = ['handout']
            elimate_overlays(content)
            eliminate_dim_reveal(content)
        else:
            doc_class_options = None

        self.document_class_obj = DocumentClass(
            document_type=backend,
            font_size=font_size,
            options=doc_class_options
        )

        if backend == 'beamer':
            # TODO: add support for custom theming, not just passing main theme str
            styles = [UseTheme(theme)]
            self.data.packages.append(
                Mode(
                    mode_type='presentation',
                    contents=styles
                )
            )

            if should_create_title_frame(title, author, date, subtitle, institutions):
                self.title_frame = TitleFrame(
                    title=title,
                    author=author,
                    date=date,
                    short_title=short_title,
                    subtitle=subtitle,
                    short_author=short_author,
                    institutions=institutions,
                    short_institution=short_institution
                )
                content.insert(0, self.title_frame)

        super().__init__(content, packages=packages, pre_env_contents=pre_env_contents)


def elimate_overlays(content):
    """
    Eliminates overlays from nested content. Modifies in place
    """
    if hasattr(content, 'content'):
        elimate_overlays(content.content)
    if hasattr(content, 'contents'):
        elimate_overlays(content.contents)
    if hasattr(content, 'overlay'):
        content.overlay = None
    if isinstance(content, (list, tuple)):
        [elimate_overlays(c) for c in content]

